function [varargout] = EKF(varargin)
%the extended Kalman filter

% Information:
% ============
% Written by: Mikael Sunnaker, mikael.sunnaker@gmail.com
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
% USA.

%Process the input
if nargin == 3,
    files = varargin{1};
    parStr = varargin{2};
    dataInfo = varargin{3}; 
else
    error('Incorrect number of input arguments');
end

%Build the matrix evaluation structure structure (ic + parameter values)
ic = parStr.ic; %initial conditions
parameters = parStr.parameters; %parameters
parNames = parStr.parNames; %parameter names
S = parStr.S; %output covariance matrix measuremnt noise
P = parStr.P; %initial covariance matrix system noise

%get the lower triangular elements of P (lower important)
lowerDP = find(tril(ones(size(P)) ~= 0));
%get the number of elements in the vectors
nrTime = length(dataInfo.tp);
nrStates = size(P,1);
%get the eye matrix
eyeMat = eye(nrStates);
%get the model
model = files.model;

%set the currently evaluated parameter values in the model
model = SBparameters(model,parNames,parameters);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%get/set the starting values for x and P
%it is assumed that the first time point for which experimenetal data
%is available is at t = 0
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%note that we sum up the noise between the first two time points as
%suggested by Overgaard et al.
simOut = TAsimulate(model,'ode23s',[dataInfo.tp(1) dataInfo.tp(2)],ic,[],files.Asym);
%the state initial conditions 
statesPred = ic(1:nrStates);
%the predicted states covariance matrix
Ppred = P;
Ppred(lowerDP) = simOut.statevalues(end,nrStates+1:end);
Ppred = Ppred + Ppred' - diag(diag(Ppred));

%loop over all points observed and saved in the vector
for i = 1:nrTime
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %Prediction
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    if(i ~= 1)
        %all states and state covariance matrix elements in one vector
        Pvect = Pud(lowerDP);
        allStates = [statesUd; Pvect];
        
        %solve the differential equation for the states
        simOut = TAsimulate(model,'ode23s',[dataInfo.tp(i-1) dataInfo.tp(i)],allStates,[],files.Asym);
        
        %get the predicted states
        statesPred = simOut.statevalues(end,1:nrStates)';
        %get the predicted covariance matrix 
        Ppred = zeros(nrStates);
        Ppred(lowerDP) = simOut.statevalues(end,nrStates+1:end);
        Ppred = Ppred + Ppred' - diag(diag(Ppred));      
    end
    
    %Calculate the Ct matrix around the predicted state values
    Ct = files.Cmat;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %Update
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %Get the output covariance matrix
    Rpred(i).mat = Ct*Ppred*Ct' + S;
    %Calculate the Kalman gain
    K = Ppred*Ct'/Rpred(i).mat;
    %Update the covariance matrix for the states, P
    %Pud = Ppred - K*Rpred(i).mat*K';
    Pud  = (eyeMat-K*Ct)*Ppred*(eyeMat-K*Ct)' + K*S*K'; %num. stable version
    
    %Calculate the residual of the output values
    ypred = statesPred(:);
    expData = dataInfo.data(i,:);
    eps(i).vect = (expData' - ypred);
    
    %Update the states
    statesUd = statesPred(1:nrStates) + K*eps(i).vect;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Function outputs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if(nargout == 2)
    varargout{1} = eps;
    varargout{2} = Rpred;
else
    error('Incorrect number of output arguments');
end

return


